import os
import glob
import pandas as pd
import math
from tqdm.auto import tqdm
import multiprocessing
import torch
from torch.utils.data import Dataset
import pickle
import gc
from contextlib import contextmanager
from filelock import FileLock
import logging, time
from typing import Optional
import joblib
from .utils import get_dict_val
from .conf import Task
nb_cores = multiprocessing.cpu_count()

@contextmanager
def disable_gc():
    gc.disable()
    try:
        yield
    finally:
        gc.enable()

class MLMDatasetOneFile(Dataset):
    def __init__(
        self,
        tokenizer,
        file_path: str,
        block_size: int,
        overwrite_cache=False,
        cache_dir: Optional[str] = None,

    ):
        assert os.path.isfile(file_path), f"Input file path {file_path} not found"

        block_size = block_size - tokenizer.num_special_tokens_to_add(pair=False)

        directory, filename = os.path.split(file_path)
        cached_features_file = os.path.join(
            cache_dir if cache_dir is not None else directory,
            "cached_lm_{}_{}_{}".format(
                tokenizer.__class__.__name__,
                str(block_size),
                filename,
            ),
        )

        # Make sure only the first process in distributed training processes the dataset,
        # and the others will use the cache.
        lock_path = cached_features_file + ".lock"
        with FileLock(lock_path):

            if os.path.exists(cached_features_file) and not overwrite_cache:
                start = time.time()
                with open(cached_features_file, "rb") as handle:
                    self.examples = joblib.load(handle)
                logging.info(
                    f"Loading features from cached file {cached_features_file} [took %.3f s]", time.time() - start
                )

            else:
                logging.info(f"Creating features from dataset file at {directory}")

                self.examples = []
                with open(file_path, encoding="utf-8") as f:
                    text = f.read()

                tokenized_text = tokenizer.convert_tokens_to_ids(tokenizer.tokenize(text))

                for i in range(0, len(tokenized_text) - block_size + 1, block_size):  # Truncate in block of block_size
                    self.examples.append(
                        tokenizer.build_inputs_with_special_tokens(tokenized_text[i : i + block_size])
                    )
                # Note that we are losing the last truncated example here for the sake of simplicity (no padding)
                # If your dataset is small, first you should loook for a bigger one :-) and second you
                # can change this behavior by adding (model specific) padding.

                start = time.time()
                with open(cached_features_file, "wb") as handle:
                    joblib.dump(self.examples, handle, protocol=pickle.HIGHEST_PROTOCOL)
                logging.info(
                    "Saving features into cached file %s [took %.3f s]", cached_features_file, time.time() - start
                )

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, i) -> torch.Tensor:
        return torch.tensor(self.examples[i], dtype=torch.long)

class MLMDataset(Dataset):
    def __init__(
        self, tokenizer, data_dir, max_length=512, binarized_path=None, ext=".txt", bs=20000,
        parallelize=True, chunksize=1_000_000, chunk_process=False
    ):
        self.fnames = glob.glob(f"{data_dir}/*{ext}")
        self.max_length = max_length
        self.tokenizer = tokenizer
        self.bs = bs
        self.features = []
        self.binarized_path = binarized_path
        self.chunksize = chunksize
        self.chunk_process = chunk_process

        if self.binarized_path is not None and self.load_binarized_features():
            assert type(self.features) == list
            if type(self.features[0]) != torch.Tensor:
                print('[INFO] Loaded data is not a list of torch.LongTensor.')
                print('[INFO] Begin converting to torch.LongTensor.\n')
                # convert on load is actually faster
                # if we try pickling torch.tensor directly
                # there are some weird too many files open bug.
                self.convert()
                print('[INFO] Done.')
        else:
            if parallelize:
                self._build_parallel(self.chunk_process)
            else:
                self._build()
                self.write_binarized_features(self.chunksize)
            self.convert()

    def __len__(self):
        return len(self.features)

    def __getitem__(self, i):
        return self.features[i]

    def _build(self):
        print('[INFO] Build features (non parallel).')
        for fname in tqdm(self.fnames):
            with open(fname, "r") as f:
                df = f.readlines()
                for i in tqdm(range(math.ceil(len(df) / self.bs))):
                    texts = list(df[i * self.bs: (i + 1) * self.bs])
                    # tokenize
                    tokenized_inputs = self.tokenizer(
                        texts,
                        max_length=self.max_length,
                        truncation=True,
                        pad_to_max_length=False
                    )
                    # add to list
                    self.features += [e for e in tokenized_inputs['input_ids']]

    def _build_one(self, fname):
        features = []
        with open(fname, "r") as f:
            df = f.readlines()
            for i in tqdm(range(math.ceil(len(df) / self.bs))):
                texts = list(df[i * self.bs: (i + 1) * self.bs])
                # tokenize
                tokenized_inputs = self.tokenizer(
                    texts,
                    max_length=self.max_length,
                    truncation=True,
                    pad_to_max_length=False
                )
                # add to list
                features += [e for e in tokenized_inputs['input_ids']]
        return features

    def _build_parallel(self, chunk_process=False):
        if chunk_process:
            print('[INFO] Build features (parallel-chunk_process).')
            # Half the core count and bumping batch size this should get
            # the same performance as small batch size but lots of cores.
            # This should save memory?
            with multiprocessing.Pool(nb_cores // 2) as pool:
                # Use imap instead this allow us to iterate on the results
                # the order is still preserved. This method save memory.
                results = pool.imap(self._build_one, self.fnames)
                start = 0
                for chunk in results:
                    self.dump_chunk(chunk, start, start + len(chunk), self.binarized_path)
                    start += len(chunk)
            print('[INFO] Start groupping results.')
            self.load_binarized_features()
            print('[INFO] Done.')
        else:
            print('[INFO] Build features (parallel).')
            with multiprocessing.Pool(nb_cores // 2) as pool:
                results = pool.map(self._build_one, self.fnames)
            print('[INFO] Start groupping results.')
            self.features = [i for lst in results for i in lst]
            print('[INFO] Done.')

    def convert(self):
        # Implementing this as multiprocessing might be slower since we will need to
        # pickle the result and sent it back later.
        with disable_gc():
            for i in range(len(self.features)):
                self.features[i] = torch.tensor(self.features[i], dtype=torch.long)
                if i % 10_000_000 == 0:
                    # Manually garbage collection.
                    gc.collect()

    @staticmethod
    def dump_chunk(data, start, stop, binarized_path):
        # Implemented as static method to pickle on chunks of data instead of entire class
        dirname = os.path.dirname(binarized_path)
        basename = os.path.basename(binarized_path)
        ext = os.path.splitext(binarized_path)[-1]
        basename = basename[:-len(ext)]
        fname = os.path.join(dirname, f'{basename}_{start}_{stop}{ext}')
        print(f'[INFO] Start writing binarized data to `{fname}`.')
        with open(fname, 'wb') as fp:
            pickle.dump(data, fp)

    def write_binarized_features(self, chunksize=None, overwrite=False):
        if self.binarized_path is not None and \
                (len(self.get_bin_fnames()) == 0 or overwrite):
            if chunksize is None:
                os.makedirs(os.path.dirname(self.binarized_path), exist_ok=True)
                print(f'[INFO] Start writing binarized data to `{self.binarized_path}`.')
                with open(self.binarized_path, 'wb') as fp:
                    pickle.dump(self.features, fp)
            else:
                os.makedirs(os.path.dirname(self.binarized_path), exist_ok=True)
                chunks = [(self.features[start: start + chunksize], start, start + chunksize,
                           self.binarized_path)
                          for start in range(0, len(self.features), chunksize)]
                # Writing list of list of int out in parallel should be fine.
                # Writing list of tensors is not working.
                with multiprocessing.Pool(nb_cores) as pool:
                    pool.starmap(self.dump_chunk, chunks)

    def _load_binarized_features(self, binarized_path):
        print(f'[INFO] Start loading binarized data from `{binarized_path}`.')
        with FileLock(lock_path):
            with open(binarized_path, 'rb') as fp:
                return pickle.load(fp)

    def load_binarized_features(self):
        print('[INFO] Load binarized data')
        bin_fnames = self.get_bin_fnames()
        if self.binarized_path is not None and \
                len(bin_fnames) > 0:
            if len(bin_fnames) == 1:
                self.features = self._load_binarized_features(bin_fnames[0])
            else:
                # This can not be parallelized for now. Since the multiprocessing
                # will need to serialize back and forth, so it will be as fast as
                # single core implementation.
                with disable_gc():
                    # Disable garbage collector for speed.
                    for i, fname in enumerate(bin_fnames):
                        self.features.extend(self._load_binarized_features(fname))
                        if (i + 1) % 10 == 0:
                            # Manually collect garbage.
                            # Do we need this?
                            gc.collect()
                            break
        return len(bin_fnames) > 0

    def get_bin_fnames(self):
        dirname = os.path.dirname(self.binarized_path)
        basename = os.path.basename(self.binarized_path)
        ext = os.path.splitext(self.binarized_path)[-1]
        basename = basename[:-len(ext)]
        bin_fnames = glob.glob(self.binarized_path) + \
            glob.glob(os.path.join(dirname, f'{basename}_*_*{ext}'))
        starts = []
        for fname in bin_fnames:
            start, _ = os.path.basename(fname).split('_')[-2:]
            starts.append(int(start))
        # Sort by starting point.
        bin_fnames = list(sorted(zip(starts, bin_fnames)))
        # Get only filenames.
        bin_fnames = [e[1] for e in bin_fnames]
        return bin_fnames


class SequenceClassificationDataset(Dataset):
    def __init__(
        self,
        tokenizer,
        data_dir,
        task=Task.MULTICLASS_CLS,
        max_length=128,
        ext=".csv",
        bs=10000,
        preprocessor=None,
        input_ids=[],
        attention_masks=[],
        labels=[],
        label_encoder=None
    ):
        self.fnames = glob.glob(f"{data_dir}/*{ext}")
        self.max_length = max_length
        self.tokenizer = tokenizer
        self.bs = bs
        self.preprocessor = preprocessor
        self.input_ids = input_ids
        self.attention_masks = attention_masks
        self.labels = labels
        self.task = task
        self.label_encoder = label_encoder
        self._build()

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, i):
        if self.task == Task.MULTICLASS_CLS:
            return {
                "input_ids": torch.tensor(self.input_ids[i], dtype=torch.long),
                "attention_mask": torch.tensor(self.attention_masks[i], dtype=torch.long),
                "label": torch.tensor(self.labels[i], dtype=torch.long),
            }
        elif self.task == Task.MULTILABEL_CLS:
            return {
                "input_ids": torch.tensor(self.input_ids[i], dtype=torch.long),
                "attention_mask": torch.tensor(self.attention_masks[i], dtype=torch.long),
                "label_ids": torch.tensor(self.labels[i], dtype=torch.float),
            }
        else:
            raise NotImplementedError

    @classmethod
    def from_dataset(cls,
                     task,
                     tokenizer,
                     dataset,
                     text_column_name,
                     label_column_name,
                     prepare_for_tokenization,
                     max_length=128,
                     bs=1000,
                     space_token='<_>',
                     preprocessor=None,
                     label_encoder=None):
        
        input_ids, attention_masks, labels = SequenceClassificationDataset._build_from_dataset(
                     task,
                     tokenizer,
                     dataset,
                     text_column_name,
                     label_column_name,
                     max_length=max_length,
                     bs=bs,
                     prepare_for_tokenization=prepare_for_tokenization,
                     space_token=space_token,
                     preprocessor=preprocessor,
                     label_encoder=label_encoder)

        return cls(
            tokenizer=tokenizer,
            data_dir=None,
            max_length=max_length,
            bs=bs,
            input_ids=input_ids,
            attention_masks=attention_masks,
            labels=labels,
            task=task,
            preprocessor=preprocessor,
            label_encoder=label_encoder
        )
    
    @staticmethod
    def _build_from_dataset(task, tokenizer, dataset,
                            text_column_name, label_column_name,
                            space_token, max_length, bs,
                            prepare_for_tokenization,
                            label_encoder,
                            preprocessor=None):
        texts = get_dict_val(dataset, text_column_name)
        if task == Task.MULTICLASS_CLS:
            labels = get_dict_val(dataset, label_column_name)

            if label_encoder != None:
                labels = label_encoder.transform(labels)

        elif task == Task.MULTILABEL_CLS:
            _labels = []
            for i, name in enumerate(label_column_name):
                # print(name)
                _labels.append(get_dict_val(dataset, name))
            labels = list(zip(*_labels))
        else:
            raise NotImplementedError
        
        input_ids = []
        attention_masks = []

        if preprocessor != None:
            print('[INFO] Apply preprocessor to texts.')
            texts = list(map(preprocessor, texts))

        for i in tqdm(range(math.ceil(len(texts) / bs))):

            batched_texts = texts[i * bs: (i+1) * bs]

            tokenized_inputs = tokenizer(
                batched_texts,
                max_length=max_length,
                truncation=True,
                # padding='max_length'            
            )
            # add to list
            input_ids += tokenized_inputs["input_ids"]
            attention_masks += tokenized_inputs["attention_mask"]
        return input_ids, attention_masks, labels

    def _build(self):
        for fname in tqdm(self.fnames):
            df = pd.read_csv(fname)
            for i in tqdm(range(math.ceil(df.shape[0] / self.bs))):
                if self.preprocessor:
                    texts = list(
                        df.iloc[i * self.bs: (i + 1) * self.bs, 0].map(
                            self.preprocessor
                        )
                    )
                else:
                    texts = list(df.iloc[i * self.bs: (i + 1) * self.bs, 0])

                # tokenize
                tokenized_inputs = self.tokenizer(
                    texts,
                    max_length=self.max_length,
                    truncation=True,
                    # padding='max_length'
                )

                # add to list
                self.input_ids += tokenized_inputs["input_ids"]
                self.attention_masks += tokenized_inputs["attention_mask"]
                self.labels += list(df.iloc[i * self.bs: (i + 1) * self.bs, 1])


class TokenClassificationDataset(Dataset):
    def __init__(
        self,
        tokenizer,
        data_dir,
        max_length=128,
        ext=".csv",
        label_pad_token="0",
        label_first_subword=False,
    ):
        self.fnames = glob.glob(f"{data_dir}/*{ext}")
        self.max_length = max_length
        self.tokenizer = tokenizer
        self.label_pad_token = label_pad_token
        self.label_first_subword = label_first_subword
        self.features = []
        self.word2sub = []

        self._build()

    def __len__(self):
        return len(self.features)

    def __getitem__(self, i):
        feature = self.features[i]
        return {
            "input_ids": torch.tensor(feature["input_ids"], dtype=torch.long),
            "attention_mask": torch.tensor(feature["attention_mask"], dtype=torch.long),
            "label": torch.tensor(feature["label"], dtype=torch.long),
            "word_ids": torch.tensor(feature["word_ids"], dtype=torch.long),
        }

    def normalize(self, src_):
        sub_1 = [
            self.tokenizer.decode(i)
            for i in self.tokenizer.encode(
                src_,
                add_special_tokens=False,
                truncation=False,
                pad_to_max_length=False,
            )
        ]

        # keep subwords only up to max length plus 3 special tokens <s>, space, </s>
        sub_2 = []
        cnt = 0
        for s in sub_1:
            if cnt > self.max_length - 3:
                break
            sub_2.append(s)
            if s != "|":
                cnt += 1

        res = "".join(sub_2)
        res = res[:-1] if res[-1] == "|" else res  # remove last space
        res = res[1:] if res[0] == "|" else res  # remove first space
        res = [" " if i == "" else i for i in res.split("|")]
        return res

    def _add_special_tokens(self, src_, lbl_):
        src = [self.tokenizer.bos_token, " "] + \
            src_ + [self.tokenizer.eos_token]
        lbl = (
            [self.label_pad_token, self.label_pad_token] +
            lbl_ + [self.label_pad_token]
        )
        txt = "".join(src)
        return src, lbl, txt

    def _get_subword_df(self, src, lbl, sub):
        # construct character df; label characters in words
        word_df = []
        for w_i in range(len(src)):
            for c in src[w_i]:
                word_df.append((c, w_i, int(lbl[w_i])))
        word_df = pd.DataFrame(word_df)
        word_df.columns = ["char", "word_i", "label"]

        # label characters in subwords
        sub_i = []
        for sw_i in range(len(sub)):
            for c in sub[sw_i]:
                sub_i.append(sw_i)
        word_df["sub_i"] = sub_i

        # map subwords to labels
        subword_df = (
            word_df.groupby(["sub_i"]).agg(
                {"label": max, "word_i": max}).reset_index()
        )

        # label for only the first subword of token
        if self.label_first_subword:
            subword_df["rnk"] = subword_df.groupby("word_i").cumcount()
            subword_df.loc[subword_df.rnk > 0, "label"] = 0

        return subword_df[["sub_i", "word_i", "label"]]

    def _build_one(self, src_, lbl_):
        # encode-decode to make sure characters are the same as in tokenizer vocab; e.g. เ เ and แ
        src_ = self.normalize(src_)
        lbl_ = lbl_.split("|")[: len(src_)]

        # # add special tokens
        src, lbl, txt = self._add_special_tokens(src_, lbl_)

        # tokenize
        tokenized_inputs = self.tokenizer(txt, add_special_tokens=False,)
        ids, attn = tokenized_inputs["input_ids"], tokenized_inputs["attention_mask"]

        # pad subwords
        to_pad = self.max_length - len(ids)
        ids += [self.tokenizer.pad_token_id] * to_pad
        attn += [0] * to_pad
        sub = [i.replace("▁", " ")
               for i in self.tokenizer.convert_ids_to_tokens(ids)]
        sub_txt = "".join(sub)

        # pad labels and words
        lbl += [self.label_pad_token] * to_pad
        src += [self.tokenizer.pad_token] * to_pad
        txt += self.tokenizer.pad_token * to_pad

        # checks
        assert len(attn) == len(ids)  # attention mask matches ids
        assert len(lbl) == len(src)  # labels match ids
        assert len(ids) == len(sub)  # ids match subwords
        assert len(txt) == len(
            sub_txt
        )  # reconstructed source text matches reconstruct subword text

        # get subword_df for metrics and matching subwords to labels
        subword_df = self._get_subword_df(src, lbl, sub)

        return {
            "input_ids": ids,
            "attention_mask": attn,
            "label": list(subword_df.label),
            "word_ids": list(subword_df.word_i),
        }

    def _build(self):
        input_ids = []
        attention_masks = []
        labels = []
        for fname in tqdm(self.fnames):
            df = pd.read_csv(fname)
            for i, row in tqdm(df.iterrows()):
                try:
                    feature = self._build_one(row[0], row[1])
                    self.features.append(feature)
                except:
                    print(row[0], row[1])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov  3 14:39:32 2020

@author: zo
"""

import warnings
import os
import multiprocessing


class REQUIRED:
    """Dummy class for checking required arguments"""
    pass


def get_field(cls, field):
    return cls.__dataclass_fields__[field]


def check_depreciated(args, warn_list):
    """Check if arguments in warn_list appear in args. Raise warning when found."""
    for arg, exp_v, warn in warn_list:
        if getattr(args, arg) != exp_v:
            warnings.warn(str(warn.args), warn.__class__)


def check_required(args):
    required = []
    kwargs = vars(args)
    for k, v in kwargs.items():
        if v == REQUIRED:
            required.append(k)
    if required:
        raise ValueError(f'{tuple(required)} are required.')


def get_file_size(f):
    """Get file size from file pointer."""
    old_file_position = f.tell()
    f.seek(0, os.SEEK_END)
    size = f.tell()
    f.seek(old_file_position, os.SEEK_SET)
    return size


def readline_clean_and_strip(file_path):
    """Generate striped line that is clean and not space from file path with utf-8 encoding"""
    with open(file_path, encoding="utf-8") as f:
        for line in _readline_clean_and_strip(f):
            yield line


def _readline_clean_and_strip(f):
    """
    Generate striped line that is clean and not space from file pointer.

    Examples::

        >>> import io
        >>> f = io.StringIO('hello\n\n\n\n  \t\t\t  world\n')
        >>> g = _readline_clean_and_strip(f)
        >>> list(g)
        ['hello', 'world']
    """
    while True:
        line = f.readline()
        if line:
            line = line.strip()
            if len(line) > 0 and not line.isspace():
                yield line
        else:
            break


def multi_imap(data, chunk_size, f,
               n_cores, progress=False):
    """
    Run function on data with multiprocessing.Pool.imap.

    Args:
        data:
            data to be separate as chunks.
        chunk_size:
            size of chunk.
        f:
            function to be execute on chunk.
        n_cores:
            number of multiprocessing cores.
        progress:
            show progress

    Returns:
        results:
            processed data.

    Examples::

        >>> import numpy
        >>> multi_imap(data=list(range(1, 5)), chunk_size=2,
                       f=numpy.exp, n_cores=2)
        [2.718281828459045, 7.38905609893065, 20.085536923187668, 54.598150033144236]
    """
    chunks = [data[i: i + chunk_size]
              for i in range(0, len(data), chunk_size)]
    with multiprocessing.Pool(n_cores) as pool:
        results = []
        for r in pool.imap(f, chunks):
            results.extend(r)
            if progress:
                print(f'\rProcessed {len(results) / len(data) * 100:.2f}%',
                      flush=True, end=' ')
    return results



import operator
from functools import reduce
from typing import Tuple, Dict, Union, List


def get_dict_val(root:Dict, keys:Union[str, List[str]]):
    """
    Access a nested object in root by item sequence.

    Args:
        root: Dict
            target object for accessing the value
        keys: Union[str, Tuple[str, str]]
            a key or a list of key (for nested structure objecy) name
            to traverse through the Dict object 

    Examples::

        >>> obj = {"a": [1,2,3]}
        >>> get_dict_val(obj, "a")
        [1,2,3]

        >>> obj = {"a": [ {"aa": 100, "bb": 0}, {"aa": 2, "bb": 5 } ] }
        >>> get_dict_val(obj, ("a", "aa"))
        [100, 2]

    """
    if type(keys) == str:
        return root[keys]
    elif type(keys) == list:
        _results = []
        for item in root[keys[0]]:
            _results.append(item[keys[1]])
        return _results
    
    return None



import numpy as np
import random
import os
import time
import mmap
import struct
import torch
from contextlib import contextmanager
from typing import Iterable

@contextmanager
def timeit(ops):
    try:
        start = time.perf_counter()
        yield
    finally:
        end = time.perf_counter()
        print(f'[{ops}] Perf: ', end - start)


@contextmanager
def callback(func):
    try:
        yield
    finally:
        func()


class MemmapIndexDataset:
    """Memory-mapped backed dataset for a list of lists of integers."""

    def __init__(self, arr_fname='arr.dat', idx_arr_fname='idx_arr.dat'):
        self.n = 0
        self.idx_arr_offset = 0
        self.arr_fname = arr_fname
        self.idx_arr_fname = idx_arr_fname
        self.mm_arr = None
        self.mm_idx_arr = None
        self.success = False

    def add(self, lstoflst: Iterable[Iterable[int]]):
        """Add list of int list to the file. This will append to the already exists data."""
        if os.path.exists(self.arr_fname) and os.path.exists(self.idx_arr_fname):
            write_mode = 'ab'
            _, _, last_arr_p, last_idx_arr_p = self.load()
        else:
            write_mode = 'wb'

        def check_sucess_if_not_revert():
            # Handle the case when user interupt the writing clear the file or revert.
            if not self.success:
                if write_mode == 'ab':
                    self.truncate(last_arr_p, last_idx_arr_p)
                    self.load()
                else:
                    self.clear()
                raise InterruptedError('Writing to buffer is not yet completed. File '
                                       'have been reverted to original or cleared.')
        with callback(check_sucess_if_not_revert):
            with open(self.arr_fname, write_mode) as arr_f, \
                 open(self.idx_arr_fname, write_mode) as idx_arr_f:
                self.success = False
                if write_mode == 'wb':
                    idx_arr_f.write(struct.pack('<I', 0))
                for lst in lstoflst:
                    length = len(lst)
                    arr_f.write(struct.pack(f'<{length}I', *lst))
                    self.idx_arr_offset += len(lst)
                    self.n += 1
                    idx_arr_f.write(struct.pack('<I', self.idx_arr_offset))
            self.success = True
            self.load()

    def load(self):
        """Load from files."""
        with open(self.arr_fname, 'r') as arr_f, \
             open(self.idx_arr_fname, 'r') as idx_arr_f:
            self.mm_arr = mmap.mmap(arr_f.fileno(), length=0, access=mmap.ACCESS_READ)
            self.mm_idx_arr = mmap.mmap(idx_arr_f.fileno(), length=0, access=mmap.ACCESS_READ)
            idx_arr_f.seek(idx_arr_f.tell(), os.SEEK_END)
            self.n = (idx_arr_f.tell() - 4) // 4
            self.idx_arr_offset, = struct.unpack('<I', self.mm_idx_arr[-4:])
            arr_f.seek(arr_f.tell(), os.SEEK_END)
            return self.n, self.idx_arr_offset, arr_f.tell(), idx_arr_f.tell()

    def truncate(self, last_arr_p, last_idx_arr_p):
        """Remove trailing data from last file pointers."""
        with open(self.arr_fname, 'r+') as arr_f, \
             open(self.idx_arr_fname, 'r+') as idx_arr_f:
            arr_f.seek(last_arr_p)
            arr_f.truncate()
            idx_arr_f.seek(last_idx_arr_p)
            idx_arr_f.truncate()

    def clear(self):
        """Remove the files."""
        if os.path.exists(self.arr_fname):
            os.remove(self.arr_fname)
        if os.path.exists(self.idx_arr_fname):
            os.remove(self.idx_arr_fname)
        self.n = 0
        self.idx_arr_offset = 0
        self.mm_arr = None
        self.mm_idx_arr = None
        self.success = False

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return tuple(self[i] for i in range(*idx.indices(len(self))))
        if idx > self.n - 1:
            raise IndexError
        elif idx < 0:
            if abs(idx) > self.n:
                raise IndexError
            return self.__getitem__(self.n + idx)
        start, stop = struct.unpack('<2I', self.mm_idx_arr[4 * idx: 4 * idx + 8])
        return struct.unpack(f'<{stop - start}I', self.mm_arr[4 * start: 4 * stop])

    def __len__(self):
        return self.n

    def __repr__(self):
        return str(self)

    def __str__(self):
        return str(tuple(tuple(e) for e in self))


class TorchMemmapIndexDataset(torch.utils.data.Dataset):

    def __init__(self, memmap_index_dataset):
        self.memmap_index_dataset = memmap_index_dataset

    def __len__(self):
        return len(self.memmap_index_dataset)

    def __getitem__(self, idx):
        return torch.LongTensor(self.memmap_index_dataset[idx])



import os
import re
from typing import List, Optional, Tuple
from shutil import copyfile
import sentencepiece as spm
import warnings
import logging
import json
import multiprocessing
from collections import Counter
from typing import Collection, Callable, Dict
from tokenizers import NormalizedString, PreTokenizedString
from transformers.tokenization_utils import PreTrainedTokenizer
from tokenizers import Tokenizer, pre_tokenizers, models
from pythainlp.tokenize import word_tokenize
from pythainlp.corpus import thai_syllables, thai_words
from pythainlp.util.trie import Trie
from functools import partial


try:
    from thai2transformers.helper import get_file_size, multi_imap
except ModuleNotFoundError:
    import sys
    sys.path.append('../scripts')  # path hacking
    from thai2transformers.helper import get_file_size, multi_imap

logger = logging.getLogger()

VOCAB_FILES_NAMES = {"vocab_file": "sentencepiece.bpe.model"}

SPIECE_UNDERLINE = '▁'
SPACE_TOKEN = "<_>"
DEPRECATED_SPACE_TOKEN = '<th_roberta_space_token>'
SEFR_SPLIT_TOKEN = '<|>'
ADDITIONAL_SPECIAL_TOKENS = ['<s>', '<pad>', '</s>', '<unk>', '<mask>', SPACE_TOKEN, '\n']
ADDITIONAL_SPECIAL_TOKENS_EXCLUDE_SPACE_TOKEN = \
    [e for e in ADDITIONAL_SPECIAL_TOKENS if e != SPACE_TOKEN]
SET_ADDITIONAL_SPECIAL_TOKENS = frozenset(ADDITIONAL_SPECIAL_TOKENS)

PRETRAINED_POSITIONAL_EMBEDDINGS_SIZES = {
    "th-roberta-base": 514,
}

# Store pre tokenizer function (text cutter)
PRE_TOKENIZERS_MAP = {'newmm': partial(
    word_tokenize,
    custom_dict=Trie(frozenset(set(thai_words()).union(set(ADDITIONAL_SPECIAL_TOKENS))))
    ),
                      'syllable': partial(
    word_tokenize,
    custom_dict=Trie(frozenset(set(thai_syllables()).union(set(ADDITIONAL_SPECIAL_TOKENS))))
    ),
    }

_nb_cores = multiprocessing.cpu_count()


def split_additional_special_token(texts):
    """
    Split list of text by additional special exclude space token.

    Args:
        texts: list of text.

    Returns:
        list_of_pre_cut_texts: list of list of pre cut text.

    Examples::

        >>> split_additional_special_token(['hello world</s></s>'])
        [['hello world', '</s>', '</s>']]
    """
    # Construct regex pattern to match additional special tokens exlude space token.
    # Not sure, if we need to escape the token but this seems to do fine.
    group = '|'.join(ADDITIONAL_SPECIAL_TOKENS_EXCLUDE_SPACE_TOKEN)
    splitter = re.compile(f'({group})')
    list_of_pre_cut_texts = []
    for text in texts:
        pre_cut_texts = []
        # Split the text this will inculde the additional token itself
        # and some time empty string in case of splitting consecutive
        # additional token.
        for e in splitter.split(text):
            # Filter out empty string space except if the string is additional
            # special token itself.
            if len(e) > 0 and (not e.isspace() or e in ADDITIONAL_SPECIAL_TOKENS):
                # Replace space token with actual space, since we want
                # to pass space into the cutter.
                pre_cut_texts.append(e.replace(SPACE_TOKEN, ' '))
        list_of_pre_cut_texts.append(pre_cut_texts)
    return list_of_pre_cut_texts


def sefr_cut_tokenize(texts, n_jobs=1, chunk_size=200):
    """
    Cut list of texts using sefr_cut.

    Args:
        texts:
            list of texts.
        n_jobs:
            Number of multiprocessing cores. -1 will use all avaliable cores.
            1 will use single core. Defaults to 1.
        chunk_size:
            size of each cutting pass in case of multiprocessing. Defaults to 200.

    Returns:
        final_list_of_cut_texts: list of list of cut text.

    Examples::
        >>> sefr_cut_tokenize(['hello world</s></s>'])
        [['hello', '<_>', 'world', '</s>', '</s>']]
    """
    if n_jobs != 1 and isinstance(texts, list):
        n_jobs = n_jobs if n_jobs != -1 else multiprocessing.cpu_count()
        return multi_imap(texts, chunk_size=chunk_size,
                          f=sefr_cut_tokenize, n_cores=n_jobs)
    if not isinstance(texts, list):
        return sefr_cut_tokenize([texts])[0]
    # We need to import the library inside the function itself to be able to use
    # multiprocessing correctly. If we did not do this. Most of the times,
    # the lock will stuck and the program will hang up.
    import sefr_cut
    import tensorflow as tf
    # Try to run tensorflow in single thread mode so we can limit the program to
    # a single process this usually give speed up with multiprocessing.
    # Because sefr_cut do tokenize each text sperately anyway. So there is not
    # much speed up to gain by using tensorflow with parallelism in conjunction
    # with multiprocessing module.
    os.environ['OMP_NUM_THREADS'] = '1'
    tf.config.threading.set_intra_op_parallelism_threads(1)
    tf.config.threading.set_inter_op_parallelism_threads(1)
    sefr_cut.load_model(engine='best')

    list_of_pre_cut_texts = split_additional_special_token(texts)
    list_of_cut_texts = []
    for pre_cut_texts in list_of_pre_cut_texts:
        cut_texts = []
        for pre_cut_text in pre_cut_texts:
            if pre_cut_text not in SET_ADDITIONAL_SPECIAL_TOKENS:
                # Tokenize pre_cut_text if it is not additional special tokens
                cut_texts.extend(sefr_cut.tokenize(pre_cut_text)[0])
            else:
                # Append token as is
                cut_texts.append(pre_cut_text)
        list_of_cut_texts.append(cut_texts)

    # Put SPACE_TOKEN back
    list_of_cut_texts = [[cut_text.replace(' ', SPACE_TOKEN) for cut_text in cut_texts]
                         for cut_texts in list_of_cut_texts]

    # Split SPACE_TOKEN out of text
    final_list_of_cut_texts = []
    splitter = re.compile(f'({SPACE_TOKEN})')
    for cut_texts in list_of_cut_texts:
        final_cut_texts = []
        for cut_text in cut_texts:
            if SPACE_TOKEN in cut_text and cut_text != SPACE_TOKEN:
                final_cut_texts.extend([e for e in splitter.split(cut_text) if len(e) > 0])
            else:
                final_cut_texts.append(cut_text)
        final_list_of_cut_texts.append(final_cut_texts)
    return final_list_of_cut_texts


# Should we do this a bit cleaner?
PRE_TOKENIZERS_MAP['sefr_cut'] = partial(sefr_cut_tokenize, n_jobs=-1)

sefr_cut_splitter = re.compile(f'({re.escape(SEFR_SPLIT_TOKEN)})')


def fake_sefr_cut_keep_split_token(text):
    """
    Split text at SEFR_SPLIT_TOKEN and kept split token.

    Args:
        text: string.

    Returns:
        list: tokens.

    Examples::

        >>> SEFR_SPLIT_TOKEN
        '<|>'
        >>> fake_sefr_cut_keep_split_token(f'hello{SEFR_SPLIT_TOKEN}world')
        ['hello', '<|>', 'world']
    """
    return [e for e in sefr_cut_splitter.split(text) if len(e) > 0]


def fake_sefr_cut(text):
    """
    Split text at SEFR_SPLIT_TOKEN.

    Args:
        text: string.

    Returns:
        list: tokens.

    Examples::

        >>> SEFR_SPLIT_TOKEN
        '<|>'
        >>> fake_sefr_cut(f'hello{SEFR_SPLIT_TOKEN}world')
        ['hello', 'world']
    """
    return text.split(SEFR_SPLIT_TOKEN)


PRE_TOKENIZERS_MAP['fake_sefr_cut'] = fake_sefr_cut
PRE_TOKENIZERS_MAP['fake_sefr_cut_keep_split_token'] = fake_sefr_cut_keep_split_token


class CustomPreTokenizer:
    def __init__(self, pre_tokenize_func: Callable):
        self.pre_tokenize_func = pre_tokenize_func

    def split(
        self, n: int, normalized_string: NormalizedString
    ) -> Collection[NormalizedString]:
        # is argument n needs?
        break_i = []
        total_i = 0
        for word in self.pre_tokenize_func(str(normalized_string)):
            total_i += len(word)
            break_i.append(total_i)
        splits = []
        last = 0
        for (i, char) in enumerate(str(normalized_string)):
            if i in break_i:
                splits.append(normalized_string[last:i])
                last = i
        splits.append(normalized_string[last:])
        return splits

    def pre_tokenize(self, pretok: PreTokenizedString):
        pretok.split(self.split)


class FakeSefrCustomTokenizer(CustomPreTokenizer):
    """
    CustomPreTokenizer that skip SEFR_SPLIT_TOKEN

    Args:
        pre_tokenizer_func: pre tokenize function.
    """

    def split(
        self, n: int, normalized_string: NormalizedString
    ) -> Collection[NormalizedString]:
        # We have to operate on original normalized string since it track aligment or something
        kept_indices = []
        p = 0
        for word in self.pre_tokenize_func(str(normalized_string)):
            if word != SEFR_SPLIT_TOKEN:
                kept_indices.append((p, p + len(word)))
            p += len(word)
        splits = []
        for start, stop in kept_indices:
            splits.append(normalized_string[start:stop])
        return splits


class WordLevelTrainer:
    """
    Trainer for word level tokenizer.

    Args:
        pre_tokenize_func:
            pre tokenize function.
        input_files:
            text files for vocabulary creation.
        additional_special_token:
            special tokens that will be explicitly added in vocabulary.
        vocab_size:
            size of vocabulary.
        vocab_min_freq:
            minimum frequency required to kept the word in vocabulary.
        progress:
            show progress.

    Examples::

        >>> trainer = WordLevelTrainer(pre_tokenize_func=pre_tokenizer_func,
                                       vocab_size=custom_args.vocab_size,
                                       vocab_min_freq=custom_args.vocab_min_freq,
                                       input_files=train_files,
                                       additional_special_tokens=additional_special_tokens)
        >>> trainer.count_parallel()
        >>> trainer.save_vocab(custom_args.output_file)
    """

    def __init__(
        self,
        pre_tokenize_func: Callable,
        input_files: str,
        additional_special_tokens: Collection[str],
        vocab_size: int = None,
        vocab_min_freq: int = None,
        progress: bool = True
    ):
        self.pre_tokenize_func = pre_tokenize_func
        self.vocab_size = vocab_size
        self.special_tokens = additional_special_tokens
        self.input_files = input_files
        self.vocab = None
        self.freq = None
        self.vocab_min_freq = vocab_min_freq
        self.progress = progress
        if self.vocab_min_freq is not None and self.vocab_size is not None:
            raise AttributeError('use only vocab_min_freq or vocab_size')

    def count_one(self, fname: str) -> Counter:
        with open(fname, "r") as f:
            file_size = get_file_size(f)
            words = []
            i = 0
            while True:
                line = f.readline()
                if line:
                    line = line.strip()
                    if len(line) > 0 and not line.isspace():
                        words.extend(self.pre_tokenize_func(line))
                else:
                    break
                i += 1
                if self.progress and i % 5000 == 0:
                    print(f'\rProcessed {f.tell() / file_size * 100:.2f}%',
                          flush=True, end=' ')
        return Counter(words)

    def count_parallel(self, nb_cores: int = _nb_cores) -> Dict[(str, int)]:
        counters = [self.count_one(fname) for fname in self.input_files]
        # disable multiprocessing for now for easier debugging
        # with multiprocessing.Pool(nb_cores) as pool:
        #     counters = pool.map(self.count_one, self.input_files)
        counter_all = sum(counters, Counter())
        # Remove special token from counter_all since this will
        # interfere with vocabulary creation later
        # for example if only '<s>' is in counter and addtional tokens = ['<s>']
        # the return vocab will be {'<s>': 1} instead of expected {'<s>': 0}
        # if we didnt remove '<s>' from counter_all
        special_tok_freq = {}
        for tok in self.special_tokens:
            if tok in counter_all:
                special_tok_freq[tok] = counter_all[tok]
                del counter_all[tok]
        if self.vocab_size is not None:
            counter_all.most_common(self.vocab_size)
        else:
            counter_all = [(key, value) for key, value in counter_all.items()
                           if value >= self.vocab_min_freq]
        self.freq = [(tok, special_tok_freq.get(tok, 0))
                     for tok in self.special_tokens] + counter_all
        self.vocab = dict((c[0], i) for i, c in enumerate(self.freq))
        return self.vocab

    def save_vocab(self, output_path: str):
        os.makedirs(os.path.dirname(output_path), exist_ok=True)
        with open(output_path, "w") as f:
            json.dump(self.vocab, f)


class ThaiRobertaTokenizer(PreTrainedTokenizer):
    """
    Adapted from :class:`~transformers.CamembertTokenizer`. Construct a
    Thai Roberta tokenizer. Based on `SentencePiece <https://github.com/google/sentencepiece>`__.

    This tokenizer inherits from :class:`~transformers.PreTrainedTokenizer` which contains most of the main methods.
    Users should refer to this superclass for more information regarding those methods.

    Args:
        vocab_file (:obj:`str`):
            `SentencePiece <https://github.com/google/sentencepiece>`__ file (generally has a `.spm` extension) that
            contains the vocabulary necessary to instantiate a tokenizer.
        bos_token (:obj:`str`, `optional`, defaults to :obj:`"<s>"`):
            The beginning of sequence token that was used during pretraining. Can be used a sequence classifier token.

            .. note::

                When building a sequence using special tokens, this is not the token that is used for the beginning
                of sequence. The token used is the :obj:`cls_token`.
        eos_token (:obj:`str`, `optional`, defaults to :obj:`"</s>"`):
            The end of sequence token.

            .. note::

                When building a sequence using special tokens, this is not the token that is used for the end
                of sequence. The token used is the :obj:`sep_token`.
        sep_token (:obj:`str`, `optional`, defaults to :obj:`"</s>"`):
            The separator token, which is used when building a sequence from multiple sequences, e.g. two sequences
            for sequence classification or for a text and a question for question answering.
            It is also used as the last token of a sequence built with special tokens.
        cls_token (:obj:`str`, `optional`, defaults to :obj:`"<s>"`):
            The classifier token which is used when doing sequence classification (classification of the whole
            sequence instead of per-token classification). It is the first token of the sequence when built with
            special tokens.
        unk_token (:obj:`str`, `optional`, defaults to :obj:`"<unk>"`):
            The unknown token. A token that is not in the vocabulary cannot be converted to an ID and is set to be this
            token instead.
        pad_token (:obj:`str`, `optional`, defaults to :obj:`"<pad>"`):
            The token used for padding, for example when batching sequences of different lengths.
        mask_token (:obj:`str`, `optional`, defaults to :obj:`"<mask>"`):
            The token used for masking values. This is the token used when training this model with masked language
            modeling. This is the token which the model will try to predict.
        additional_special_tokens (:obj:`List[str]`, `optional`, defaults to :obj:`["<s>NOTUSED", "</s>NOTUSED"]`):
            Additional special tokens used by the tokenizer.

    Attributes:
        sp_model (:obj:`SentencePieceProcessor`):
            The `SentencePiece` processor that is used for every conversion (string, tokens and IDs).
    """

    vocab_files_names = VOCAB_FILES_NAMES
    max_model_input_sizes = PRETRAINED_POSITIONAL_EMBEDDINGS_SIZES
    model_input_names = ["attention_mask"]

    def __init__(
        self,
        vocab_file,
        bos_token="<s>",
        eos_token="</s>",
        sep_token="</s>",
        cls_token="<s>",
        unk_token="<unk>",
        pad_token="<pad>",
        mask_token="<mask>",
        additional_special_tokens=[SPACE_TOKEN],
        **kwargs
    ):
        super().__init__(
            bos_token=bos_token,
            eos_token=eos_token,
            unk_token=unk_token,
            sep_token=sep_token,
            cls_token=cls_token,
            pad_token=pad_token,
            mask_token=mask_token,
            additional_special_tokens=additional_special_tokens,
            **kwargs,
        )
        self.sp_model = spm.SentencePieceProcessor()
        self.sp_model.Load(str(vocab_file))
        self.vocab_file = vocab_file

    def build_inputs_with_special_tokens(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        """
        Build model inputs from a sequence or a pair of sequence for sequence classification tasks
        by concatenating and adding special tokens.
        An CamemBERT sequence has the following format:

        - single sequence: ``<s> X </s>``
        - pair of sequences: ``<s> A </s></s> B </s>``

        Args:
            token_ids_0 (:obj:`List[int]`):
                List of IDs to which the special tokens will be added.
            token_ids_1 (:obj:`List[int]`, `optional`):
                Optional second list of IDs for sequence pairs.

        Returns:
            :obj:`List[int]`: List of `input IDs <../glossary.html#input-ids>`__ with the appropriate special tokens.
        """

        if token_ids_1 is None:
            return [self.cls_token_id] + token_ids_0 + [self.sep_token_id]
        cls = [self.cls_token_id]
        sep = [self.sep_token_id]
        return cls + token_ids_0 + sep + sep + token_ids_1 + sep

    def get_special_tokens_mask(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None, already_has_special_tokens: bool = False
    ) -> List[int]:
        """
        Retrieve sequence ids from a token list that has no special tokens added. This method is called when adding
        special tokens using the tokenizer ``prepare_for_model`` method.

        Args:
            token_ids_0 (:obj:`List[int]`):
                List of IDs.
            token_ids_1 (:obj:`List[int]`, `optional`):
                Optional second list of IDs for sequence pairs.
            already_has_special_tokens (:obj:`bool`, `optional`, defaults to :obj:`False`):
                Whether or not the token list is already formatted with special tokens for the model.

        Returns:
            :obj:`List[int]`: A list of integers in the range [0, 1]: 1 for a special token, 0 for a sequence token.
        """
        if already_has_special_tokens:
            if token_ids_1 is not None:
                raise ValueError(
                    "You should not supply a second sequence if the provided sequence of "
                    "ids is already formated with special tokens for the model."
                )
            return list(map(lambda x: 1 if x in [self.sep_token_id, self.cls_token_id] else 0, token_ids_0))

        if token_ids_1 is None:
            return [1] + ([0] * len(token_ids_0)) + [1]
        return [1] + ([0] * len(token_ids_0)) + [1, 1] + ([0] * len(token_ids_1)) + [1]

    def create_token_type_ids_from_sequences(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        """
        Create a mask from the two sequences passed to be used in a sequence-pair classification task.
        CamemBERT, like RoBERTa, does not make use of token type ids, therefore a list of zeros is returned.

        Args:
            token_ids_0 (:obj:`List[int]`):
                List of IDs.
            token_ids_1 (:obj:`List[int]`, `optional`):
                Optional second list of IDs for sequence pairs.

        Returns:
            :obj:`List[int]`: List of zeros.
        """
        sep = [self.sep_token_id]
        cls = [self.cls_token_id]

        if token_ids_1 is None:
            return len(cls + token_ids_0 + sep) * [0]
        return len(cls + token_ids_0 + sep + sep + token_ids_1 + sep) * [0]

    @property
    def vocab_size(self):
        return len(self.sp_model)

    def get_vocab(self):
        vocab = {self.convert_ids_to_tokens(i): i for i in range(self.vocab_size)}
        return vocab

    def _tokenize(self, text):
        return self.sp_model.EncodeAsPieces(text)

    def _convert_token_to_id(self, token):
        """ Converts a token (str) in an id using the vocab. """
        return self.sp_model.PieceToId(token)

    def _convert_id_to_token(self, index):
        """Converts an index (integer) in a token (str) using the vocab."""
        return self.sp_model.IdToPiece(index)

    def __getstate__(self):
        state = self.__dict__.copy()
        state["sp_model"] = None
        return state

    def __setstate__(self, d):
        self.__dict__ = d
        self.sp_model = spm.SentencePieceProcessor()
        self.sp_model.Load(self.vocab_file)

    def convert_tokens_to_string(self, tokens):
        """Converts a sequence of tokens (strings for sub-words) in a single string."""
        out_string = "".join(tokens).replace(SPIECE_UNDERLINE, "\n").strip()
        return out_string

    def save_vocabulary(self, save_directory: str, filename_prefix: Optional[str] = None) -> Tuple[str]:
        if not os.path.isdir(save_directory):
            logger.error("Vocabulary path ({}) should be a directory".format(save_directory))
            return
        out_vocab_file = os.path.join(
            save_directory, (filename_prefix + "-" if filename_prefix else "") + VOCAB_FILES_NAMES["vocab_file"]
        )

        if os.path.abspath(self.vocab_file) != os.path.abspath(out_vocab_file):
            copyfile(self.vocab_file, out_vocab_file)

        return (out_vocab_file,)

    def prepare_for_tokenization(self, text, space_token=SPACE_TOKEN, is_split_into_words=False, **kwargs):
        if "is_pretokenized" in kwargs:
            warnings.warn(
                "`is_pretokenized` is deprecated and will be removed in a future version, use `is_split_into_words` instead.",
                FutureWarning,
            )
            is_split_into_words = kwargs.pop("is_pretokenized")

        # replace empty space with special space token

        text = text.replace(' ', space_token)

        return (text, kwargs)


class BaseThaiWordsTokenizer(PreTrainedTokenizer):
    """Base cass for word level tokenizer."""

    def build_inputs_with_special_tokens(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        """
        Build model inputs from a sequence or a pair of sequence for sequence classification tasks
        by concatenating and adding special tokens.
        An CamemBERT sequence has the following format:

        - single sequence: ``<s> X </s>``
        - pair of sequences: ``<s> A </s></s> B </s>``

        Args:
            token_ids_0 (:obj:`List[int]`):
                List of IDs to which the special tokens will be added.
            token_ids_1 (:obj:`List[int]`, `optional`):
                Optional second list of IDs for sequence pairs.

        Returns:
            :obj:`List[int]`: List of `input IDs <../glossary.html#input-ids>`__ with the appropriate special tokens.
        """
        if token_ids_1 is None:
            return [self.cls_token_id] + token_ids_0 + [self.sep_token_id]
        cls = [self.cls_token_id]
        sep = [self.sep_token_id]
        return cls + token_ids_0 + sep + sep + token_ids_1 + sep

    def get_special_tokens_mask(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None, already_has_special_tokens: bool = False
    ) -> List[int]:
        """
        Retrieve sequence ids from a token list that has no special tokens added. This method is called when adding
        special tokens using the tokenizer ``prepare_for_model`` method.

        Args:
            token_ids_0 (:obj:`List[int]`):
                List of IDs.
            token_ids_1 (:obj:`List[int]`, `optional`):
                Optional second list of IDs for sequence pairs.
            already_has_special_tokens (:obj:`bool`, `optional`, defaults to :obj:`False`):
                Whether or not the token list is already formatted with special tokens for the model.

        Returns:
            :obj:`List[int]`: A list of integers in the range [0, 1]: 1 for a special token, 0 for a sequence token.
        """
        if already_has_special_tokens:
            if token_ids_1 is not None:
                raise ValueError(
                    "You should not supply a second sequence if the provided sequence of "
                    "ids is already formated with special tokens for the model."
                )
            return list(map(lambda x: 1 if x in [self.sep_token_id, self.cls_token_id] else 0, token_ids_0))

        if token_ids_1 is None:
            return [1] + ([0] * len(token_ids_0)) + [1]
        return [1] + ([0] * len(token_ids_0)) + [1, 1] + ([0] * len(token_ids_1)) + [1]

    def create_token_type_ids_from_sequences(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        """
        Create a mask from the two sequences passed to be used in a sequence-pair classification task.
        CamemBERT, like RoBERTa, does not make use of token type ids, therefore a list of zeros is returned.

        Args:
            token_ids_0 (:obj:`List[int]`):
                List of IDs.
            token_ids_1 (:obj:`List[int]`, `optional`):
                Optional second list of IDs for sequence pairs.

        Returns:
            :obj:`List[int]`: List of zeros.
        """
        sep = [self.sep_token_id]
        cls = [self.cls_token_id]

        if token_ids_1 is None:
            return len(cls + token_ids_0 + sep) * [0]
        return len(cls + token_ids_0 + sep + sep + token_ids_1 + sep) * [0]

    @property
    def vocab_size(self):
        return len(self.tokenizer_model.get_vocab())

    def get_vocab(self):
        vocab = {self.convert_ids_to_tokens(i): i for i in range(self.vocab_size)}
        return vocab

    def _tokenize(self, text):
        return self.tokenizer_model.encode(text).tokens

    def _convert_token_to_id(self, token):
        """ Converts a token (str) in an id using the vocab. """
        i = self.tokenizer_model.token_to_id(token)
        if i is None:
            return self.unk_token_id
        else:
            return i

    def _convert_id_to_token(self, index):
        """Converts an index (integer) in a token (str) using the vocab."""
        return self.tokenizer_model.id_to_token(index)

    def convert_tokens_to_string(self, tokens):
        """Converts a sequence of tokens (strings for sub-words) in a single string."""
        out_string = "".join(tokens).strip()
        return out_string

    def save_vocabulary(self, save_directory: str, filename_prefix: Optional[str] = None) -> Tuple[str]:
        if not os.path.isdir(save_directory):
            logger.error("Vocabulary path ({}) should be a directory".format(save_directory))
            return
        out_vocab_file = os.path.join(
            save_directory, (filename_prefix + "-" if filename_prefix else "") + VOCAB_FILES_NAMES["vocab_file"]
        )

        if os.path.abspath(self.vocab_file) != os.path.abspath(out_vocab_file):
            copyfile(self.vocab_file, out_vocab_file)

        return (out_vocab_file,)

    def prepare_for_tokenization(self, text, space_token=SPACE_TOKEN, is_split_into_words=False, **kwargs):
        if "is_pretokenized" in kwargs:
            warnings.warn(
                "`is_pretokenized` is deprecated and will be removed in a future version, use `is_split_into_words` instead.",
                FutureWarning,
            )
            is_split_into_words = kwargs.pop("is_pretokenized")

        # replace empty space with special space token

        text = text.replace(' ', space_token)

        return (text, kwargs)

    def __getstate__(self):
        # What is this funcion even do?
        raise NotImplementedError

    def __setstate__(self, d):
        # What is this funcion even do?
        raise NotImplementedError


class ThaiWordsNewmmTokenizer(BaseThaiWordsTokenizer):
    """
    Newmm tokenizer.
    """
    vocab_files_names = {"vocab_file": "newmm.json"}  # vocabulary file location in folder

    def __init__(
        self,
        vocab_file,
        bos_token="<s>",
        eos_token="</s>",
        sep_token="</s>",
        cls_token="<s>",
        unk_token="<unk>",
        pad_token="<pad>",
        mask_token="<mask>",
        additional_special_tokens=ADDITIONAL_SPECIAL_TOKENS,
        **kwargs
    ):
        super().__init__(
            bos_token=bos_token,
            eos_token=eos_token,
            unk_token=unk_token,
            sep_token=sep_token,
            cls_token=cls_token,
            pad_token=pad_token,
            mask_token=mask_token,
            additional_special_tokens=ADDITIONAL_SPECIAL_TOKENS,
            **kwargs,
        )
        pre_tokenizer_func = PRE_TOKENIZERS_MAP['newmm']
        custom_pre_tokenizer = pre_tokenizers.PreTokenizer.custom(
            CustomPreTokenizer(pre_tokenizer_func))
        tokenizer = Tokenizer(models.WordLevel.from_file(vocab_file))
        tokenizer.pre_tokenizer = custom_pre_tokenizer
        self.tokenizer_model = tokenizer
        self.vocab_file = vocab_file

    def __getstate__(self):
        state = self.__dict__.copy()
        state["tokenizer_model"] = None
        return state

    def __setstate__(self, d):
        self.__dict__ = d
        pre_tokenizer_func = PRE_TOKENIZERS_MAP['newmm']
        custom_pre_tokenizer = pre_tokenizers.PreTokenizer.custom(
            CustomPreTokenizer(pre_tokenizer_func))
        tokenizer = Tokenizer(models.WordLevel.from_file(self.vocab_file))
        tokenizer.pre_tokenizer = custom_pre_tokenizer
        self.tokenizer_model = tokenizer


class ThaiWordsSyllableTokenizer(BaseThaiWordsTokenizer):
    """
    Syllable tokenizer.
    """
    vocab_files_names = {"vocab_file": "syllable.json"}

    def __init__(
        self,
        vocab_file,
        bos_token="<s>",
        eos_token="</s>",
        sep_token="</s>",
        cls_token="<s>",
        unk_token="<unk>",
        pad_token="<pad>",
        mask_token="<mask>",
        additional_special_tokens=ADDITIONAL_SPECIAL_TOKENS,
        **kwargs
    ):
        super().__init__(
            bos_token=bos_token,
            eos_token=eos_token,
            unk_token=unk_token,
            sep_token=sep_token,
            cls_token=cls_token,
            pad_token=pad_token,
            mask_token=mask_token,
            additional_special_tokens=ADDITIONAL_SPECIAL_TOKENS,
            **kwargs,
        )
        pre_tokenizer_func = PRE_TOKENIZERS_MAP['syllable']
        custom_pre_tokenizer = pre_tokenizers.PreTokenizer.custom(
            CustomPreTokenizer(pre_tokenizer_func))
        tokenizer = Tokenizer(models.WordLevel.from_file(vocab_file))
        tokenizer.pre_tokenizer = custom_pre_tokenizer
        self.tokenizer_model = tokenizer
        self.vocab_file = vocab_file

    def __getstate__(self):
        state = self.__dict__.copy()
        state["tokenizer_model"] = None
        return state

    def __setstate__(self, d):
        self.__dict__ = d
        pre_tokenizer_func = PRE_TOKENIZERS_MAP['syllable']
        custom_pre_tokenizer = pre_tokenizers.PreTokenizer.custom(
            CustomPreTokenizer(pre_tokenizer_func))
        tokenizer = Tokenizer(models.WordLevel.from_file(self.vocab_file))
        tokenizer.pre_tokenizer = custom_pre_tokenizer
        self.tokenizer_model = tokenizer


class FakeSefrCutTokenizer(BaseThaiWordsTokenizer):
    """
    FakeSefrCut tokenizer.
    """
    vocab_files_names = {"vocab_file": "fake_sefr_cut.json"}

    def __init__(
        self,
        vocab_file,
        bos_token="<s>",
        eos_token="</s>",
        sep_token="</s>",
        cls_token="<s>",
        unk_token="<unk>",
        pad_token="<pad>",
        mask_token="<mask>",
        additional_special_tokens=ADDITIONAL_SPECIAL_TOKENS,
        **kwargs
    ):
        super().__init__(
            bos_token=bos_token,
            eos_token=eos_token,
            unk_token=unk_token,
            sep_token=sep_token,
            cls_token=cls_token,
            pad_token=pad_token,
            mask_token=mask_token,
            additional_special_tokens=ADDITIONAL_SPECIAL_TOKENS,
            **kwargs,
        )
        pre_tokenizer_func = PRE_TOKENIZERS_MAP['fake_sefr_cut_keep_split_token']
        custom_pre_tokenizer = pre_tokenizers.PreTokenizer.custom(
            FakeSefrCustomTokenizer(pre_tokenizer_func))
        tokenizer = Tokenizer(models.WordLevel.from_file(vocab_file))
        tokenizer.pre_tokenizer = custom_pre_tokenizer
        self.tokenizer_model = tokenizer
        self.vocab_file = vocab_file



# -*- coding: utf-8 -*-
"""
Preprocessing for thai2transformers
"""
from typing import Collection, Callable
from functools import partial
import re
import html
import emoji
from pythainlp.tokenize import word_tokenize


_TK_UNK, _TK_REP, _TK_WREP, _TK_URL, _TK_END = "<unk> <rep> <wrep> <url> </s>".split()

SPACE_SPECIAL_TOKEN = "<_>"

# str->str rules
def fix_html(text: str) -> str:
    """
        List of replacements from html strings in `test`. (code from `fastai`)
        :param str text: text to replace html string
        :return: text where html strings are replaced
        :rtype: str
        :Example:
            >>> fix_html("Anbsp;amp;nbsp;B @.@ ")
            A & B.
    """
    re1 = re.compile(r"  +")
    text = (
        text.replace("#39;", "'")
        .replace("amp;", "&")
        .replace("#146;", "'")
        .replace("nbsp;", " ")
        .replace("#36;", "$")
        .replace("\\n", "\n")
        .replace("quot;", "'")
        .replace("<br />", "\n")
        .replace('\\"', '"')
        .replace(" @.@ ", ".")
        .replace(" @-@ ", "-")
        .replace(" @,@ ", ",")
        .replace("\\", " \\ ")
    )
    return re1.sub(" ", html.unescape(text))


def replace_url(text: str) -> str:
    """
        Replace url in `text` with TK_URL (https://stackoverflow.com/a/6041965)
        :param str text: text to replace url
        :return: text where urls  are replaced
        :rtype: str
        :Example:
            >>> replace_url("go to https://github.com")
            go to <url>
    """
    URL_PATTERN = r"(http|ftp|https)://([\w_-]+(?:(?:\.[\w_-]+)+))([\w.,@?^=%&:/~+#-]*[\w@?^=%&/~+#-])?"
    return re.sub(URL_PATTERN, _TK_URL, text)


def rm_brackets(text: str) -> str:
    """
        Remove all empty brackets and artifacts within brackets from `text`.
        :param str text: text to remove useless brackets
        :return: text where all useless brackets are removed
        :rtype: str
        :Example:
            >>> rm_brackets("hey() whats[;] up{*&} man(hey)")
            hey whats up man(hey)
    """
    # remove empty brackets
    new_line = re.sub(r"\(\)", "", text)
    new_line = re.sub(r"\{\}", "", new_line)
    new_line = re.sub(r"\[\]", "", new_line)
    # brakets with only punctuations
    new_line = re.sub(r"\([^a-zA-Z0-9ก-๙]+\)", "", new_line)
    new_line = re.sub(r"\{[^a-zA-Z0-9ก-๙]+\}", "", new_line)
    new_line = re.sub(r"\[[^a-zA-Z0-9ก-๙]+\]", "", new_line)
    # artifiacts after (
    new_line = re.sub(r"(?<=\()[^a-zA-Z0-9ก-๙]+(?=[a-zA-Z0-9ก-๙])", "", new_line)
    new_line = re.sub(r"(?<=\{)[^a-zA-Z0-9ก-๙]+(?=[a-zA-Z0-9ก-๙])", "", new_line)
    new_line = re.sub(r"(?<=\[)[^a-zA-Z0-9ก-๙]+(?=[a-zA-Z0-9ก-๙])", "", new_line)
    # artifacts before )
    new_line = re.sub(r"(?<=[a-zA-Z0-9ก-๙])[^a-zA-Z0-9ก-๙]+(?=\))", "", new_line)
    new_line = re.sub(r"(?<=[a-zA-Z0-9ก-๙])[^a-zA-Z0-9ก-๙]+(?=\})", "", new_line)
    new_line = re.sub(r"(?<=[a-zA-Z0-9ก-๙])[^a-zA-Z0-9ก-๙]+(?=\])", "", new_line)
    return new_line


def replace_newlines(text: str) -> str:
    """
        Replace newlines in `text` with spaces.
        :param str text: text to replace all newlines with spaces
        :return: text where all newlines are replaced with spaces
        :rtype: str
        :Example:
            >>> rm_useless_spaces("hey whats\n\nup")
            hey whats  up
    """

    return re.sub(r"[\n]", " ", text.strip())


def rm_useless_spaces(text: str) -> str:
    """
        Remove multiple spaces in `text`. (code from `fastai`)
        :param str text: text to replace useless spaces
        :return: text where all spaces are reduced to one
        :rtype: str
        :Example:
            >>> rm_useless_spaces("oh         no")
            oh no
    """
    return re.sub(" {2,}", " ", text)

def replace_spaces(text: str, space_token: str = SPACE_SPECIAL_TOKEN) -> str:
    """
        Replace spaces with _
        :param str text: text to replace spaces
        :return: text where all spaces replaced with _
        :rtype: str
        :Example:
            >>> replace_spaces("oh no")
            oh_no
    """
    return re.sub(" ", space_token, text)


def replace_rep_after(text: str) -> str:
    """
    Replace repetitions at the character level in `text`
    :param str text: input text to replace character repetition
    :return: text with repetitive tokens removed.
    :rtype: str
    :Example:
        >>> text = "กาาาาาาา"
        >>> replace_rep_after(text)
        'กา'
    """

    def _replace_rep(m):
        c, cc = m.groups()
        return f"{c}"

    re_rep = re.compile(r"(\S)(\1{3,})")
    return re_rep.sub(_replace_rep, text)


# List[str] -> List[str] rules
def ungroup_emoji(toks: Collection[str]) -> Collection[str]:
    """
    Ungroup Zero Width Joiner (ZVJ) Emojis
    See https://emojipedia.org/emoji-zwj-sequence/
    :param Collection[str] toks: list of tokens
    :return: list of tokens where emojis are ungrouped
    :rtype: Collection[str]
    :Example:
        >>> toks = []
        >>> ungroup_emoji(toks)
        []
    """
    res = []
    for tok in toks:
        if emoji.emoji_count(tok) == len(tok):
            res.extend(list(tok))
        else:
            res.append(tok)
    return res


def replace_wrep_post(toks: Collection[str]) -> Collection[str]:
    """
    Replace reptitive words post tokenization;
    fastai `replace_wrep` does not work well with Thai.
    :param Collection[str] toks: list of tokens
    :return: list of tokens where repetitive words are removed.
    :rtype: Collection[str]
    :Example:
        >>> toks = ["กา", "น้ำ", "น้ำ", "น้ำ", "น้ำ"]
        >>> replace_wrep_post(toks)
        ['กา', 'น้ำ']
    """
    previous_word = None
    rep_count = 0
    res = []
    for current_word in toks + [_TK_END]:
        if current_word == previous_word:
            rep_count += 1
        elif (current_word != previous_word) & (rep_count > 0):
            res += [previous_word]
            rep_count = 0
        else:
            res.append(previous_word)
        previous_word = current_word
    return res[1:]


def remove_space(toks: Collection[str]) -> Collection[str]:
    """
    Do not include space for bag-of-word models.
    :param Collection[str] toks: list of tokens
    :return: Collection of tokens where space tokens (" ") are filtered out
    :rtype: Collection[str]
    :Example:
        >>> toks = ['ฉัน','เดิน',' ','กลับ','บ้าน']
        >>> remove_space(toks)
        ['ฉัน','เดิน','กลับ','บ้าน']
    """
    res = []
    for t in toks:
        t = t.strip()
        if t:
            res.append(t)
    return res


# combine them together
def process_transformers(
    text: str,
    pre_rules: Collection[Callable] = [
        fix_html,
        rm_brackets,
        replace_newlines,
        rm_useless_spaces,
        replace_spaces,
        replace_rep_after,
    ],
    tok_func: Callable = word_tokenize,
    post_rules: Collection[Callable] = [ungroup_emoji, replace_wrep_post],
) -> str:
    text = text.lower()
    for rule in pre_rules:
        text = rule(text)
    toks = tok_func(text)
    for rule in post_rules:
        toks = rule(toks)
    return "".join(toks)



from collections import OrderedDict

import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader


from thai2transformers.metrics import classification_metrics
from thai2transformers.datasets import (
    SequenceClassificationDataset,
    TokenClassificationDataset
)
from transformers import (
    AdamW,
    get_linear_schedule_with_warmup,
    AutoTokenizer,
    AutoModel,
    AutoConfig
)
from transformers.modeling_outputs import (
    SequenceClassifierOutput
)

from transformers.modeling_bert import (
    BertPreTrainedModel,
    BertModel,
    BertConfig
)

from transformers.modeling_roberta import (
    RobertaPreTrainedModel,
    RobertaModel,
    RobertaConfig,
    RobertaClassificationHead
)

from transformers.modeling_xlm_roberta import (
    XLMRobertaConfig
)

class BertForMultiLabelSequenceClassification(BertPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels

        self.bert = BertModel(config)
        self.dropout = torch.nn.Dropout(config.hidden_dropout_prob)
        self.classifier = torch.nn.Linear(config.hidden_size, config.num_labels)
        self.init_weights()

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        labels=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=True,
    ):
        r"""
        labels (:obj:`torch.LongTensor` of shape :obj:`(batch_size,)`, `optional`):
            Labels for computing the sequence classification/regression loss.
            Indices should be in :obj:`[0, ..., config.num_labels - 1]`.
            If :obj:`config.num_labels > 1` a multilabel classification loss is computed (BCELoss).
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        outputs = self.bert(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        pooled_output = outputs[1]

        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)

        loss = None
        if labels is not None:
            loss_fct = torch.nn.BCEWithLogitsLoss()
            loss = loss_fct(logits, labels)

        if not return_dict:
            output = (logits,) + outputs[2:]
            return ((loss,) + output) if loss is not None else output

        return SequenceClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )


class RobertaForMultiLabelSequenceClassification(RobertaPreTrainedModel):
    authorized_missing_keys = [r"position_ids"]

    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels

        self.roberta = RobertaModel(config, add_pooling_layer=False)
        self.classifier = RobertaClassificationHead(config)
        self.init_weights()  

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        labels=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=True,
    ):
        r"""
        labels (:obj:`torch.LongTensor` of shape :obj:`(batch_size,)`, `optional`):
            Labels for computing the sequence classification/regression loss.
            Indices should be in :obj:`[0, ..., config.num_labels - 1]`.
            If :obj:`config.num_labels > 1` a multilabel classification loss is computed (BCELoss).
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        outputs = self.roberta(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        sequence_output = outputs[0]
        logits = self.classifier(sequence_output)

        loss = None
        if labels is not None:
            if self.num_labels > 1:
                loss_fct = torch.nn.BCEWithLogitsLoss()
                loss = loss_fct(logits, labels)

        if not return_dict:
            output = (logits,) + outputs[2:]
            return ((logits,) + output) if loss is not None else output

        return SequenceClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )

class XLMRobertaForMultiLabelSequenceClassification(RobertaForMultiLabelSequenceClassification):
    """
    This class overrides :class:`~thai2transformers.RobertaForMultiLabelSequenceClassificationn`. Please check the
    superclass for the appropriate documentation alongside usage examples.
    """

    config_class = XLMRobertaConfig



# -*- coding: utf-8 -*-
from setuptools import find_packages, setup

with open('README.md', 'r') as f:
  readme =  f.read()

requirements = [
    "torch>=1.5.0",
    "transformers==3.5.0",
    "sentencepiece==0.1.94",
    "emoji",
    "pythainlp>=2.2.4",
    "tqdm",
    "datasets>=1.2.1",
    "sefr_cut",
    "seqeval",
    "pytorch-lightning",
    "pydantic",
    "jsonlines",
    "ssg==0.0.6",
    "tokenizers==0.9.4",
    "pandas==1.1.4",
]


setup(
    name="thai2transformers",
    version="0.1.0",
    description="Pretraining transformer based Thai language models",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="VISTEC-depa AI Research Institute of Thailand",
    author_email="",
    url="https://github.com/vistec-AI/thai2transformers",
    packages=find_packages(exclude=["notebooks", "notebooks.*", "external_scripts", "external_scripts.*"]),
    python_requires=">=3.6",
    package_data={},
    include_package_data=True,
    install_requires=requirements,
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords=[
        "thainlp",
        "NLP",
        "natural language processing",
        "text analytics",
        "text processing",
        "localization",
        "computational linguistics",
        "ThaiNLP",
        "Thai NLP",
        "Thai language",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: Thai",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Text Processing",
        "Topic :: Text Processing :: General",
        "Topic :: Text Processing :: Linguistic",
    ],
    project_urls={
        "Documentation": "https://github.com/vistec-AI/thai2transformers",
        "Tutorials": "https://colab.research.google.com/drive/1Kbk6sBspZLwcnOE61adAQo30xxqOQ9ko?usp=sharing",
        "Source Code": "https://github.com/vistec-AI/thai2transformers",
        "Bug Tracker": "https://github.com/vistec-AI/thai2transformers/issues",
    },
)



import pandas as pd
import numpy as np
from sklearn.metrics import f1_score, accuracy_score, precision_recall_fscore_support, classification_report
from seqeval.metrics import (accuracy_score as seqeval_accuracy_score, 
                             classification_report as seqeval_classification_report, 
                             f1_score as seqeval_f1_score,
                             precision_score as seqeval_precision_score, 
                             recall_score as seqeval_recall_score)

def sk_classification_metrics(pred, pred_labs=False):
    result = classification_metrics(pred)
    labels = pred.label_ids
    preds = pred.predictions if pred_labs else pred.predictions.argmax(-1) 
    result['classification_report'] = classification_report(labels, preds, digits=4)
    return result


def classification_metrics(pred, pred_labs=False):
    labels = pred.label_ids
    preds = pred.predictions if pred_labs else pred.predictions.argmax(-1)
    precision_macro, recall_macro, f1_macro, _ = precision_recall_fscore_support(labels, preds, average="macro")
    precision_micro, recall_micro, f1_micro, _ = precision_recall_fscore_support(labels, preds, average="micro")
    acc = accuracy_score(labels, preds)
    return {
        'accuracy': acc,
        'f1_micro': f1_micro,
        'precision_micro': precision_micro,
        'recall_micro': recall_micro,
        'f1_macro': f1_macro,
        'precision_macro': precision_macro,
        'recall_macro': recall_macro,
        'nb_samples': len(labels)
    }


def seqeval_classification_metrics(pred):
    from seqeval.metrics import accuracy_score, precision_score, recall_score, f1_score, classification_report
    labels = pred.label_ids
    preds = pred.predictions
    precision_macro = precision_score(labels, preds, average='macro')
    recall_macro = recall_score(labels, preds, average='macro')
    f1_macro = f1_score(labels, preds, average='macro')
    precision_micro = precision_score(labels, preds, average='micro')
    recall_micro = recall_score(labels, preds, average='micro')
    f1_micro = f1_score(labels, preds, average='micro')
    acc = accuracy_score(labels, preds)
    return {
        'accuracy': acc,
        'f1_micro': f1_micro,
        'precision_micro': precision_micro,
        'recall_micro': recall_micro,
        'f1_macro': f1_macro,
        'precision_macro': precision_macro,
        'recall_macro': recall_macro,
        'nb_samples': len(labels),
        'classification_report': classification_report(labels, preds, digits=4)
    }

def _compute_best_threshold(targets, probs):
    f1s = []
    for threshold in range(1,100):
        preds = (probs > (threshold / 100)).astype(int)
        f1s.append((
            threshold/100,
            f1_score(targets,
                     preds,
                     average='binary')
        ))

    f1s_df = pd.DataFrame(f1s).sort_values(1,ascending=False).reset_index(drop=True)
    f1s_df.columns = ['threshold_label','f1_label']

    return f1s_df.threshold_label[0], f1s_df.f1_label[0]

def _select_best_thresholds(targets, probs, n_labels):
    best_thresholds = dict()
    for i in range(0, n_labels):
        
        best_thresholds[f'label-{i}'] = _compute_best_threshold(targets[:,i], probs[:,i])
        
    return best_thresholds

def sigmoid(x):
    return 1/(1 + np.exp(-x)) 

def multilabel_classification_metrics(pred, n_labels):

    labels = pred.label_ids
    logits = pred.predictions
    probs = sigmoid(logits)

    best_threshold_mapping = _select_best_thresholds(labels, probs, n_labels)
    best_thresholds = [ v[0] for k,v in best_threshold_mapping.items() ]

    preds = np.array(probs > best_thresholds)
    precision_macro, recall_macro, f1_macro, _ = precision_recall_fscore_support(labels, preds, average='macro')
    precision_micro, recall_micro, f1_micro, _ = precision_recall_fscore_support(labels, preds, average='micro')
    acc = accuracy_score(labels, preds)
    accuracy_micro = (labels == preds).mean()

    return {
        'accuracy': acc,
        'accuracy_micro': accuracy_micro,
        'f1_micro': f1_micro,
        'precision_micro': precision_micro,
        'recall_micro': recall_micro,
        'f1_macro': f1_macro,
        'precision_macro': precision_macro,
        'recall_macro': recall_macro,
        'nb_samples': len(labels)
    }



import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader
import pytorch_lightning as pl
import argparse
from thai2transformers.metrics import classification_metrics
from thai2transformers.datasets import (
    SequenceClassificationDataset,
    TokenClassificationDataset,
)
from transformers import (
    AdamW,
    get_linear_schedule_with_warmup,
    AutoTokenizer,
    AutoModel,
)

class TokenClassificationFinetuner(pl.LightningModule):
    def __init__(self, hparams):
        super(TokenClassificationFinetuner, self).__init__()

        self.hparams = hparams
        self.model = AutoModel.from_pretrained(self.hparams.model_name_or_path,)
        self.tokenizer = AutoTokenizer.from_pretrained(self.hparams.model_name_or_path,)
        self.head = torch.nn.Sequential(
            torch.nn.Linear(self.hparams.num_hidden, self.hparams.num_hidden),
            torch.nn.Dropout(p=self.hparams.drop_p),
            torch.nn.Linear(self.hparams.num_hidden, self.hparams.num_labels),
        )
        self.loss_fn = torch.nn.CrossEntropyLoss()

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
    ):
        o1, o2 = self.model(
            input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
        )
        preds = self.head(o1)
        return preds

    def _flatten_batch(self, batch):
        preds = self.forward(
            input_ids=batch["input_ids"], attention_mask=batch["attention_mask"],
        )

        active_mask = batch["attention_mask"].view(-1) == 1
        active_preds = preds.view(-1, self.hparams.num_labels)
        active_labels = torch.where(
            active_mask,
            batch["label"].view(-1),
            torch.tensor(self.loss_fn.ignore_index).type_as(batch["label"]),
        )
        return active_preds, active_labels, preds

    def _step(self, batch):
        active_preds, active_labels, preds = self._flatten_batch(batch)
        loss = self.loss_fn(active_preds, active_labels)
        return loss, preds

    def training_step(self, batch, batch_nb):
        loss, _ = self._step(batch)
        tensorboard_logs = {"train_loss": loss}
        return {"loss": loss, "log": tensorboard_logs}

    def validation_step(self, batch, batch_nb):
        loss, preds = self._step(batch)
        pred_labs = preds.argmax(2).cpu().numpy()
        
        #get batch labels and pred_labs
        dfs = []
        for i in range(batch['input_ids'].shape[0]):
            df = pd.DataFrame({'word_ids':batch['word_ids'][i].cpu().numpy(),
                               'label':batch['label'][i].cpu().numpy(),
                               'pred_labs':pred_labs[i]})
            df = df[df.label!=0].groupby('word_ids').max().reset_index(drop=True)
            dfs.append(df)
        df_batch = pd.concat(dfs)
        
        #get metrics
        pred = argparse.Namespace(
            label_ids=df_batch.label, predictions=df_batch.pred_labs,
        )
        results = classification_metrics(pred, pred_labs=True)
        results["loss"] = loss.cpu()
        return results

    def test_step(self, batch, batch_nb):
        return self.validation_step(batch, batch_nb)
    
    def _avg_epoch_end(self, outputs):
        total_samples = np.sum([x["nb_samples"] for x in outputs])
        avg_loss = np.sum([x["loss"] * x["nb_samples"] for x in outputs]) / total_samples
        avg_acc = np.sum([x["accuracy"] * x["nb_samples"] for x in outputs]) / total_samples
        #micro
        avg_f1_micro = np.sum([x["f1_micro"] * x["nb_samples"] for x in outputs]) / total_samples
        avg_precision_micro = np.sum(
            [x["precision_micro"] * x["nb_samples"] for x in outputs]
        ) / total_samples
        avg_recall_micro = np.sum([x["recall_micro"] * x["nb_samples"] for x in outputs]) / total_samples
        #macro
        avg_f1_macro = np.sum([x["f1_macro"] * x["nb_samples"] for x in outputs]) / total_samples
        avg_precision_macro = np.sum(
            [x["precision_macro"] for x in outputs]
        ) / total_samples
        avg_recall_macro = np.sum([x["recall_macro"] * x["nb_samples"] for x in outputs]) / total_samples
        
        return {
            "avg_loss": avg_loss,
            "avg_acc": avg_acc,
            "avg_f1_micro": avg_f1_micro,
            "avg_precision_micro": avg_precision_micro,
            "avg_recall_micro": avg_recall_micro,
            "avg_f1_macro": avg_f1_macro,
            "avg_precision_macro": avg_precision_macro,
            "avg_recall_macro": avg_recall_macro,
            "total_samples": total_samples,
        }
    
    def validation_epoch_end(self, outputs):
        results = self._avg_epoch_end(outputs)
        tensorboard_logs = {
            "val_loss": results["avg_loss"],
            "avg_val_acc": results["avg_acc"],
            "avg_val_f1_micro": results["avg_f1_micro"],
            "avg_val_precision_micro": results["avg_precision_micro"],
            "avg_val_recall_micro": results["avg_recall_micro"],
            "avg_val_f1_macro": results["avg_f1_macro"],
            "avg_val_precision_macro": results["avg_precision_macro"],
            "avg_val_recall_macro": results["avg_recall_macro"],
            "total_samples": results["total_samples"]
        }
        return {
            "val_loss": results["avg_loss"],
            "log": tensorboard_logs,
            "progress_bar": tensorboard_logs,
        }

    def test_epoch_end(self, outputs):
        results = self._avg_epoch_end(outputs)
        tensorboard_logs = {
            "test_loss": results["avg_loss"],
            "avg_test_acc": results["avg_acc"],
            "avg_test_f1_micro": results["avg_f1_micro"],
            "avg_test_precision_micro": results["avg_precision_micro"],
            "avg_test_recall_micro": results["avg_recall_micro"],
            "avg_test_f1_macro": results["avg_f1_macro"],
            "avg_test_precision_macro": results["avg_precision_macro"],
            "avg_test_recall_macro": results["avg_recall_macro"],
            "total_samples": results["total_samples"]
        }
        return {
            "test_loss": results["avg_loss"],
            "log": tensorboard_logs,
            "progress_bar": tensorboard_logs,
        }

    def configure_optimizers(self):
        model = self.model
        no_decay = ["bias", "LayerNorm.weight", "LayerNorm.weight"]
        optimizer_grouped_parameters = [
            {
                "params": [
                    p
                    for n, p in model.named_parameters()
                    if not any(nd in n for nd in no_decay)
                ],
                "weight_decay": self.hparams.weight_decay,
            },
            {
                "params": [
                    p
                    for n, p in model.named_parameters()
                    if any(nd in n for nd in no_decay)
                ],
                "weight_decay": 0.0,
            },
        ]
        optimizer = AdamW(
            optimizer_grouped_parameters,
            lr=self.hparams.learning_rate,
            eps=self.hparams.adam_epsilon,
        )
        self.opt = optimizer
        return [optimizer]

    def optimizer_step(
        self,
        epoch,
        batch_idx,
        optimizer,
        optimizer_idx,
        second_order_closure=None,
        on_tpu=False,
        using_native_amp=False,
        using_lbfgs=False,
    ):
        optimizer.step()
        optimizer.zero_grad()
        self.lr_scheduler.step()

    def train_dataloader(self):
        train_dataset = TokenClassificationDataset(
            tokenizer=self.tokenizer,
            data_dir=self.hparams.train_dir,
            max_length=self.hparams.max_length,
            label_pad_token=self.hparams.label_pad_token,
            label_first_subword=self.hparams.label_first_subword,
        )

        dataloader = DataLoader(
            train_dataset,
            batch_size=self.hparams.per_device_train_batch_size,
            drop_last=True,
            shuffle=True,
            num_workers=4,
        )

        # calculate total timesteps
        t_total = (
            (
                len(dataloader.dataset)
                // (
                    self.hparams.per_device_train_batch_size
                    * max(1, self.hparams.n_gpu)
                )
            )
            // self.hparams.gradient_accumulation_steps
            * float(self.hparams.num_train_epochs)
        )
        # create scheduler
        scheduler = get_linear_schedule_with_warmup(
            self.opt,
            num_warmup_steps=self.hparams.warmup_steps,
            num_training_steps=t_total,
        )
        self.lr_scheduler = scheduler
        return dataloader

    def val_dataloader(self):
        val_dataset = TokenClassificationDataset(
            tokenizer=self.tokenizer,
            data_dir=self.hparams.valid_dir,
            max_length=self.hparams.max_length,
            label_pad_token=self.hparams.label_pad_token,
            label_first_subword=self.hparams.label_first_subword,
        )

        return DataLoader(
            val_dataset,
            batch_size=self.hparams.per_device_eval_batch_size,
            num_workers=4,
        )

    def test_dataloader(self):
        test_dataset = TokenClassificationDataset(
            tokenizer=self.tokenizer,
            data_dir=self.hparams.test_dir,
            max_length=self.hparams.max_length,
            label_pad_token=self.hparams.label_pad_token,
            label_first_subword=self.hparams.label_first_subword,
        )

        return DataLoader(
            test_dataset,
            batch_size=self.hparams.per_device_eval_batch_size,
            num_workers=4,
        )


class SequenceClassificationFinetuner(pl.LightningModule):
    def __init__(self, hparams):
        super(SequenceClassificationFinetuner, self).__init__()

        self.hparams = hparams
        self.model = AutoModel.from_pretrained(self.hparams.model_name_or_path,)
        self.tokenizer = AutoTokenizer.from_pretrained(self.hparams.model_name_or_path,)
        self.head = torch.nn.Sequential(
            torch.nn.Linear(self.hparams.num_hidden, self.hparams.num_hidden),
            torch.nn.Dropout(p=self.hparams.drop_p),
            torch.nn.Linear(self.hparams.num_hidden, self.hparams.num_labels),
        )
        self.loss_fn = torch.nn.CrossEntropyLoss()

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
    ):
        o1, o2 = self.model(
            input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
        )
        preds = self.head(o2)
        return preds

    def _step(self, batch):
        preds = self.forward(
            input_ids=batch["input_ids"], attention_mask=batch["attention_mask"],
        )

        loss = self.loss_fn(preds, batch["label"])

        return loss, preds

    def training_step(self, batch, batch_nb):
        loss, _ = self._step(batch)
        tensorboard_logs = {"train_loss": loss}
        return {"loss": loss, "log": tensorboard_logs}

    def validation_step(self, batch, batch_nb):
        loss, preds = self._step(batch)
        labels = batch["label"]
        pred = argparse.Namespace(label_ids=labels.cpu(), predictions=preds.cpu())
        results = classification_metrics(pred)
        results["loss"] = loss.cpu()
        return results

    def test_step(self, batch, batch_nb):
        return self.validation_step(batch, batch_nb)

    def _avg_epoch_end(self, outputs):
        total_samples = np.sum([x["nb_samples"] for x in outputs])
        avg_loss = np.sum([x["loss"] * x["nb_samples"] for x in outputs]) / total_samples
        avg_acc = np.sum([x["accuracy"] * x["nb_samples"] for x in outputs]) / total_samples
        #micro
        avg_f1_micro = np.sum([x["f1_micro"] * x["nb_samples"] for x in outputs]) / total_samples
        avg_precision_micro = np.sum(
            [x["precision_micro"] * x["nb_samples"] for x in outputs]
        ) / total_samples
        avg_recall_micro = np.sum([x["recall_micro"] * x["nb_samples"] for x in outputs]) / total_samples
        #macro
        avg_f1_macro = np.sum([x["f1_macro"] * x["nb_samples"] for x in outputs]) / total_samples
        avg_precision_macro = np.sum(
            [x["precision_macro"] for x in outputs]
        ) / total_samples
        avg_recall_macro = np.sum([x["recall_macro"] * x["nb_samples"] for x in outputs]) / total_samples
        
        return {
            "avg_loss": avg_loss,
            "avg_acc": avg_acc,
            "avg_f1_micro": avg_f1_micro,
            "avg_precision_micro": avg_precision_micro,
            "avg_recall_micro": avg_recall_micro,
            "avg_f1_macro": avg_f1_macro,
            "avg_precision_macro": avg_precision_macro,
            "avg_recall_macro": avg_recall_macro,
            "total_samples": total_samples,
        }
    
    def validation_epoch_end(self, outputs):
        results = self._avg_epoch_end(outputs)
        tensorboard_logs = {
            "val_loss": results["avg_loss"],
            "avg_val_acc": results["avg_acc"],
            "avg_val_f1_micro": results["avg_f1_micro"],
            "avg_val_precision_micro": results["avg_precision_micro"],
            "avg_val_recall_micro": results["avg_recall_micro"],
            "avg_val_f1_macro": results["avg_f1_macro"],
            "avg_val_precision_macro": results["avg_precision_macro"],
            "avg_val_recall_macro": results["avg_recall_macro"],
            "total_samples": results["total_samples"]
        }
        return {
            "val_loss": results["avg_loss"],
            "log": tensorboard_logs,
            "progress_bar": tensorboard_logs,
        }

    def test_epoch_end(self, outputs):
        results = self._avg_epoch_end(outputs)
        tensorboard_logs = {
            "test_loss": results["avg_loss"],
            "avg_test_acc": results["avg_acc"],
            "avg_test_f1_micro": results["avg_f1_micro"],
            "avg_test_precision_micro": results["avg_precision_micro"],
            "avg_test_recall_micro": results["avg_recall_micro"],
            "avg_test_f1_macro": results["avg_f1_macro"],
            "avg_test_precision_macro": results["avg_precision_macro"],
            "avg_test_recall_macro": results["avg_recall_macro"],
            "total_samples": results["total_samples"]
        }
        return {
            "test_loss": results["avg_loss"],
            "log": tensorboard_logs,
            "progress_bar": tensorboard_logs,
        }

    def configure_optimizers(self):
        model = self.model
        no_decay = ["bias", "LayerNorm.weight", "LayerNorm.weight"]
        optimizer_grouped_parameters = [
            {
                "params": [
                    p
                    for n, p in model.named_parameters()
                    if not any(nd in n for nd in no_decay)
                ],
                "weight_decay": self.hparams.weight_decay,
            },
            {
                "params": [
                    p
                    for n, p in model.named_parameters()
                    if any(nd in n for nd in no_decay)
                ],
                "weight_decay": 0.0,
            },
        ]
        optimizer = AdamW(
            optimizer_grouped_parameters,
            lr=self.hparams.learning_rate,
            eps=self.hparams.adam_epsilon,
        )
        self.opt = optimizer
        return [optimizer]

    def optimizer_step(
        self,
        epoch,
        batch_idx,
        optimizer,
        optimizer_idx,
        second_order_closure=None,
        on_tpu=False,
        using_native_amp=False,
        using_lbfgs=False,
    ):
        optimizer.step()
        optimizer.zero_grad()
        self.lr_scheduler.step()

    def train_dataloader(self):
        train_dataset = SequenceClassificationDataset(
            tokenizer=self.tokenizer,
            data_dir=self.hparams.train_dir,
            max_length=self.hparams.max_length,
        )

        dataloader = DataLoader(
            train_dataset,
            batch_size=self.hparams.per_device_train_batch_size,
            drop_last=True,
            shuffle=True,
            num_workers=4,
        )

        # calculate total timesteps
        t_total = (
            (
                len(dataloader.dataset)
                // (
                    self.hparams.per_device_train_batch_size
                    * max(1, self.hparams.n_gpu)
                )
            )
            // self.hparams.gradient_accumulation_steps
            * float(self.hparams.num_train_epochs)
        )
        # create scheduler
        scheduler = get_linear_schedule_with_warmup(
            self.opt,
            num_warmup_steps=self.hparams.warmup_steps,
            num_training_steps=t_total,
        )
        self.lr_scheduler = scheduler
        return dataloader

    def val_dataloader(self):
        val_dataset = SequenceClassificationDataset(
            tokenizer=self.tokenizer,
            data_dir=self.hparams.valid_dir,
            max_length=self.hparams.max_length,
        )

        return DataLoader(
            val_dataset,
            batch_size=self.hparams.per_device_eval_batch_size,
            num_workers=4,
        )

    def test_dataloader(self):
        test_dataset = SequenceClassificationDataset(
            tokenizer=self.tokenizer,
            data_dir=self.hparams.test_dir,
            max_length=self.hparams.max_length,
        )

        return DataLoader(
            test_dataset,
            batch_size=self.hparams.per_device_eval_batch_size,
            num_workers=4,
        )



from collections import OrderedDict

from transformers import (
    AutoConfig,
    PretrainedConfig
)

from transformers.modeling_bert import BertConfig
from transformers.modeling_roberta import RobertaConfig
from transformers.modeling_xlm_roberta import XLMRobertaConfig

from .models import (
    XLMRobertaForMultiLabelSequenceClassification,
    BertForMultiLabelSequenceClassification,
    RobertaForMultiLabelSequenceClassification
)

MODEL_FOR_MULTI_LABEL_SEQUENCE_CLASSIFICATION_MAPPING = OrderedDict(
    [
        (XLMRobertaConfig, XLMRobertaForMultiLabelSequenceClassification),
        (BertConfig, BertForMultiLabelSequenceClassification),        
        (RobertaConfig, RobertaForMultiLabelSequenceClassification),
    ]
)

class AutoModelForMultiLabelSequenceClassification:
    
    def __init__(self):
        raise EnvironmentError(
            "AutoModelForMultiLabelSequenceClassification is designed to be instantiated "
            "using the `AutoModelForMultiLabelSequenceClassification.from_pretrained(pretrained_model_name_or_path)` or "
            "`AutoModelForMultiLabelSequenceClassification.from_config(config)` methods."
        )

    @classmethod
    def from_config(cls, config):
        r"""
        Instantiates one of the model classes of the library---with a sequence classification head---from a
        configuration.

        Note:
            Loading a model from its configuration file does **not** load the model weights.
            It only affects the model's configuration. Use
            :meth:`~transformers.AutoModelForMultiLabelSequenceClassification.from_pretrained` to load the model weights.

        Args:
            config (:class:`~transformers.PretrainedConfig`):
                The model class to instantiate is selected based on the configuration class:

                List options

        Examples::

            >>> from transformers import AutoConfig, AutoModelForMultiLabelSequenceClassification
            >>> # Download configuration from S3 and cache.
            >>> config = AutoConfig.from_pretrained('bert-base-uncased')
            >>> model = AutoModelForMultiLabelSequenceClassification.from_config(config)
        """
        if type(config) in MODEL_FOR_MULTI_LABEL_SEQUENCE_CLASSIFICATION_MAPPING.keys():
            return MODEL_FOR_MULTI_LABEL_SEQUENCE_CLASSIFICATION_MAPPING[type(config)](config)
        raise ValueError(
            "Unrecognized configuration class {} for this kind of AutoModel: {}.\n"
            "Model type should be one of {}.".format(
                config.__class__,
                cls.__name__,
                ", ".join(c.__name__ for c in MODEL_FOR_MULTI_LABEL_SEQUENCE_CLASSIFICATION_MAPPING.keys()),
            )
        )

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, *model_args, **kwargs):
        r"""
        Examples::

            >>> from transformers import AutoConfig, AutoModelForMultiLabelSequenceClassification

            >>> # Download model and configuration from S3 and cache.
            >>> model = AutoModelForMultiLabelSequenceClassification.from_pretrained('bert-base-uncased')

            >>> # Update configuration during loading
            >>> model = AutoModelForMultiLabelSequenceClassification.from_pretrained('bert-base-uncased', output_attentions=True)
            >>> model.config.output_attentions
            True

            >>> # Loading from a TF checkpoint file instead of a PyTorch model (slower)
            >>> config = AutoConfig.from_json_file('./tf_model/bert_tf_model_config.json')
            >>> model = AutoModelForMultiLabelSequenceClassification.from_pretrained('./tf_model/bert_tf_checkpoint.ckpt.index', from_tf=True, config=config)
        """
        config = kwargs.pop("config", None)
        if not isinstance(config, PretrainedConfig):
            config, kwargs = AutoConfig.from_pretrained(
                pretrained_model_name_or_path, return_unused_kwargs=True, **kwargs
            )

        if type(config) in MODEL_FOR_MULTI_LABEL_SEQUENCE_CLASSIFICATION_MAPPING.keys():
            return MODEL_FOR_MULTI_LABEL_SEQUENCE_CLASSIFICATION_MAPPING[type(config)].from_pretrained(
                pretrained_model_name_or_path, *model_args, config=config, **kwargs
            )
        raise ValueError(
            "Unrecognized configuration class {} for this kind of AutoModel: {}.\n"
            "Model type should be one of {}.".format(
                config.__class__,
                cls.__name__,
                ", ".join(c.__name__ for c in MODEL_FOR_MULTI_LABEL_SEQUENCE_CLASSIFICATION_MAPPING.keys()),
            )
        )
